-------------------------------------------------------------------------------
-- Title      : Higher-level quad ADC controller for ADC128S022
-- Project    : 
-------------------------------------------------------------------------------
-- File       : adc.vhdl
-- Author     : <kwalsh@holycross.edu>
-- Company    : 
-- Created    : 2016-02-07
-- Last update: 2016-02-07
-- Platform   : 
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: On startup, the controller immediately begins, and indefinitely
--              continues, reading the first four inputs of the ADC.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author      Description
-- 2016-02-07  1.0      kwalsh      Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity adc4 is


  port (
        clk            : in  std_logic; -- Clock must be between 0.8 MHz and 3.2 MHz.

	-- Channel data is available approx. 16*4 cycles after startup and is
	-- updated continuously thereafter.
        chan0          : out std_logic_vector(15 downto 0);
        chan1          : out std_logic_vector(15 downto 0);
        chan2          : out std_logic_vector(15 downto 0);
        chan3          : out std_logic_vector(15 downto 0);

        adc_saddr      : out std_logic;
        adc_cs_n       : out std_logic;
        adc_sclk       : out std_logic;
        adc_data       : in  std_logic
  );

end adc4;

architecture rtl of adc4 is

  signal ready : std_logic := '0';

  signal cn : std_logic_vector(10 downto 0) := "00000000000";
  signal c0 : std_logic_vector(11 downto 0) := X"000";
  signal c1 : std_logic_vector(11 downto 0) := X"000";
  signal c2 : std_logic_vector(11 downto 0) := X"000";
  signal c3 : std_logic_vector(11 downto 0) := X"000";

  signal nxt : std_logic_vector(2 downto 0) := "001";
  signal cnt : integer range 0 to 15 := 0;

begin

  chan0 <= X"0" & c0;
  chan1 <= X"0" & c1;
  chan2 <= X"0" & c2;
  chan3 <= X"0" & c3;

  adc_cs_n <= '0' when ready = '1' else '1';
  adc_sclk <= clk when ready = '1' else '1';

  --    clk __/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\
  --  ready __/'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  --  cs_n  ''\_______________________________________________________________________________________________________________________________________
  --  sclk  '''''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\__/''\
  --  saddr -----< dc  X dc  X a2  X a1  X a0  >-----------------------------------------------------------------<  dc X  dc X  a2 X a1  X a0  >------
  --   data -----< dc  X dc  X dc  X dc  X 11  X 10  X 9   X 8   X 7   X 6   X 5   X 4   X 3   X 2   X 1   X 0   X dc  X dc  X dc  X dc  X 11  X 10  X 
  --    cnt 0    X 1   X 2   X 3   X 4   X 5   X 6   X 7   X 8   X 9   X 10  X 11  X 12  X 13  X 14  X 15  X 0   X 1   X 2   X 3   X 4   X 5   X 6   X 

  process (clk)
  begin

    if rising_edge(clk) then
      if ready = '0' then
	ready <= '1';
      else
	if cnt > 4 then
	  cn <= cn(9 downto 0) & adc_data; -- shift/capture
	elsif cnt = 0 then
	  if nxt = "001" then
	    c0 <= cn(10 downto 0) & adc_data; -- final shift/capture, transfer to dest
	    nxt <= "010";
	  elsif nxt = "010" then
	    c1 <= cn(10 downto 0) & adc_data; -- final shift/capture, transfer to dest
	    nxt <= "011";
	  elsif nxt = "011" then
	    c2 <= cn(10 downto 0) & adc_data; -- final shift/capture, transfer to dest
	    nxt <= "000";
	  elsif nxt = "000" then
	    c3 <= cn(10 downto 0) & adc_data; -- final shift/capture, transfer to dest
	    nxt <= "001";
	  end if;
	  cn <= "00000000000";
	end if; -- last count
      end if; -- ready
    end if; -- rising edges

    if falling_edge(clk) and ready = '1' then
      if cnt = 2 then
	adc_saddr <= nxt(2);
      elsif cnt = 3 then
	adc_saddr <= nxt(1);
      elsif cnt = 4 then
	adc_saddr <= nxt(0);
      else
	adc_saddr <= '0';
      end if;

      if cnt = 15 then
	cnt <= 0;
      else
	cnt <= cnt + 1;
      end if;
    end if; -- falling edges

  end process;

end architecture rtl;
