-------------------------------------------------------------------------------
-- Title      : Low-level 4-bit LCD controller
-- Project    : 
-------------------------------------------------------------------------------
-- File       : lcdq4.vhdl
-- Author     : <kwalsh@holycross.edu>
-- Company    : 
-- Created    : 2016-02-07
-- Last update: 2016-02-07
-- Platform   : 
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: The controller initializes the LCD display, then waits for
--              commands and data on the inputs and transmits them over the
--              4-bit LCD bus.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author      Description
-- 2016-02-07  1.0      kwalsh      Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity lcdq4 is

--  generic (CLK_PERIOD_NS : positive := 20);    -- 50MHz

  port (
        clk50                 : in  std_logic;
        rst                   : in  std_logic;
        cmd                   : in  std_logic_vector(7 downto 0);
        data                  : in  std_logic_vector(7 downto 0);
        status                : out std_logic_vector(7 downto 0);
        write                 : in  std_logic;
        read                  : in  std_logic;
        chipselect            : in  std_logic_vector(15 downto 0);
        lcd_db4567_en_rw_rs   : out std_logic_vector(6 downto 0)
  );

end lcdq4;

architecture rtl of lcdq4 is

  constant CLK_PERIOD_NS : positive := 20; -- 50MHz
  constant DELAY_15_MS   : positive := 15 * 10**6 / CLK_PERIOD_NS + 1;
  constant DELAY_1640_US : positive := 1640 * 10**3 / CLK_PERIOD_NS + 1;
  constant DELAY_4100_US : positive := 4100 * 10**3 / CLK_PERIOD_NS + 1;
  constant DELAY_100_US  : positive := 100 * 10**3 / CLK_PERIOD_NS + 1;
  constant DELAY_40_US   : positive := 40 * 10**3 / CLK_PERIOD_NS + 1;

  constant DELAY_NIBBLE     : positive := 10**3 / CLK_PERIOD_NS + 1;
  constant DELAY_LCD_E      : positive := 230 / CLK_PERIOD_NS + 1;
  constant DELAY_SETUP_HOLD : positive := 40 / CLK_PERIOD_NS + 1;

  constant MAX_DELAY : positive := DELAY_15_MS;

  -- this record describes one write operation
  type op_t is record
    valid   : std_logic;
    seq     : std_logic;
    rs      : std_logic;
    data    : std_logic_vector(7 downto 0);
    delay_h : integer range 0 to MAX_DELAY;
    delay_l : integer range 0 to MAX_DELAY;
  end record op_t;
  constant nop : op_t := (valid => '0', seq => '0', rs => '0', data => X"00", delay_h => 0, delay_l => 0);

  type state_t is (
    SEND_RESET,
    SEND_CONFIG,
    SEND_READY
  );

  signal send_state      : state_t := SEND_RESET;
  signal send_next_state : state_t;

  -- init + config operations:
  -- write 3 x 0x3 followed by 0x2
  -- function set command
  -- entry mode set command
  -- display on/off command
  -- clear display
  type config_ops_t is array(0 to 12) of op_t;
  constant config_ops : config_ops_t := (
    12 => (valid => '1', seq => '0', rs => '0', data => X"33", delay_h => DELAY_4100_US, delay_l => DELAY_100_US), -- 0x3, 0x3
    11 => (valid => '1', seq => '0', rs => '0', data => X"32", delay_h => DELAY_40_US, delay_l => DELAY_40_US), -- 0x3, 0x2
    10 => (valid => '1', seq => '0', rs => '0', data => X"28", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- function_set(4bit, 2line, 6x8font) -- was 0x24
    9 => (valid => '1', seq => '0', rs => '0', data => X"06", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- entry_mode(incr, noshift) -- was 0x04
    8 => (valid => '1', seq => '0', rs => '0', data => X"0C", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- display(on, nocursor, noblink) -- was 0x0f
    7 => (valid => '1', seq => '0', rs => '0', data => X"01", delay_h => DELAY_NIBBLE, delay_l => DELAY_1640_US), -- clear
    6 => (valid => '1', seq => '0', rs => '0', data => X"80", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- setaddr(0)
    5 => (valid => '1', seq => '0', rs => '1', data => X"52", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- setdata('A')
    4 => (valid => '1', seq => '0', rs => '1', data => X"65", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- setdata('B')
    3 => (valid => '1', seq => '0', rs => '1', data => X"61", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- setdata('C')
    2 => (valid => '1', seq => '0', rs => '1', data => X"64", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- setdata('C')
    1 => (valid => '1', seq => '0', rs => '1', data => X"79", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US), -- setdata('C')
    0 => (valid => '1', seq => '0', rs => '1', data => X"21", delay_h => DELAY_NIBBLE, delay_l => DELAY_40_US)  -- setdata('C')
  );

  signal send_cnt      : natural range 0 to 12 := 0;
  signal send_next_cnt : natural range 0 to 12;


  signal send_next_op : op_t;
  signal xmit_op : op_t := nop;
  signal xmit_ack : std_logic := '0';

  type xmit_state_t is (
    XMIT_IDLE,
    XMIT_SETUP_H, XMIT_ENABLE_H, XMIT_HOLD_H, XMIT_DELAY_H,
    XMIT_SETUP_L, XMIT_ENABLE_L, XMIT_HOLD_L, XMIT_DELAY_L
  );

  signal xmit_state      : xmit_state_t := XMIT_IDLE;
  signal xmit_next_state : xmit_state_t;
  signal xmit_cnt        : natural range 0 to MAX_DELAY;
  signal xmit_next_cnt   : natural range 0 to MAX_DELAY;

begin

  -- send fsm: starts in SEND_RESET; does SEND_CONFIG a few times to initialize
  -- LCD, then stays in SEND_READY indefinitely.
  -- Timing of state transitions is controlled by xmit_ack = xmit_op.seq.

  send_pump : process (clk50) is
  begin
    if rising_edge(clk50) then
      if rst = '1' then
        send_state <= SEND_RESET;
        send_cnt   <= 0;
        xmit_op    <= nop;
      elsif xmit_ack = xmit_op.seq then
        send_state <= send_next_state;
        send_cnt   <= send_next_cnt;
        xmit_op    <= send_next_op;
	if send_next_op.valid = '1' then
	  xmit_op.seq <= not xmit_op.seq;
	end if;
      end if;
    end if;
  end process send_pump;

  send_fsm : process(send_state, send_cnt, cmd, data, read, write, chipselect) is
  begin
    case send_state is
      when SEND_RESET =>
        send_next_op    <= nop;
        send_next_state <= SEND_CONFIG;
        send_next_cnt   <= config_ops_t'high;

      when SEND_CONFIG =>
        send_next_op <= config_ops(send_cnt);
        if send_cnt = 0 then
          send_next_state <= SEND_READY;
          send_next_cnt   <= 0;
        else
          send_next_state <= SEND_CONFIG;
          send_next_cnt   <= send_cnt - 1;
        end if;

      when SEND_READY =>
        send_next_state <= SEND_READY;
        send_next_cnt   <= 0;
        if write = '1' and chipselect = X"0001" then -- register 0, one hot encoding
	  send_next_op.valid   <= '1';
	  send_next_op.seq     <= '0';
	  send_next_op.rs      <= '0';
          send_next_op.data    <= cmd;
          send_next_op.delay_h <= DELAY_NIBBLE;
          send_next_op.delay_l <= DELAY_40_US;
        elsif write = '1' and chipselect = X"0002" then -- register 1, one hot encoding
	  send_next_op.valid   <= '1';
	  send_next_op.seq     <= '0';
          send_next_op.rs      <= '1';
          send_next_op.data    <= data;
          send_next_op.delay_h <= DELAY_NIBBLE;
          send_next_op.delay_l <= DELAY_40_US;
        else
          send_next_op <= nop;
        end if;

    end case;
  end process send_fsm;


  -- xmit_fsm: starts in XMIT_IDLE; when xmit_op.valid with a new xmit_op.seq
  -- number is detected, transitions through several states to transmit the
  -- 8-bit xmit_op.data over the 4-bit LCD bus. Timing of state transitions is
  -- controlled by xmit_cnt.

  xmit_pump : process (clk50) is
  begin
    if rising_edge(clk50) then
      if rst = '1' then
	xmit_ack   <= '0';
        xmit_state <= XMIT_IDLE;
        xmit_cnt   <= 0;
      elsif xmit_cnt = 0 then
	if xmit_next_state = XMIT_IDLE then
	  xmit_ack <= xmit_op.seq;
	else
	  xmit_ack <= xmit_ack;
	end if;
        xmit_state <= xmit_next_state;
        xmit_cnt   <= xmit_next_cnt;
      else
	xmit_ack   <= xmit_ack;
	xmit_state <= xmit_state;
        xmit_cnt   <= xmit_cnt - 1;
      end if;
    end if;
  end process xmit_pump;

  xmit_fsm : process(xmit_state, xmit_op, xmit_ack) is
  begin
    case xmit_state is

      -- idle: wait for xmit_op.valid and different xmit_op.seq
      when XMIT_IDLE =>
        lcd_db4567_en_rw_rs <= X"0" & '0' & '0' & '0';
        if xmit_op.valid = '0' or xmit_op.seq = xmit_ack then
          xmit_next_state <= XMIT_IDLE;
          xmit_next_cnt   <= 0;
        else
          xmit_next_state <= XMIT_SETUP_H;
          xmit_next_cnt   <= DELAY_SETUP_HOLD;
        end if;

      -- xmit high nibble: setup, enable, hold, delay
      when XMIT_SETUP_H =>
        lcd_db4567_en_rw_rs <= xmit_op.data(4) & xmit_op.data(5) & xmit_op.data(6) & xmit_op.data(7) & '0' & '0' & xmit_op.rs;
        xmit_next_state <= XMIT_ENABLE_H;
        xmit_next_cnt   <= DELAY_LCD_E;

      when XMIT_ENABLE_H =>
        lcd_db4567_en_rw_rs <= xmit_op.data(4) & xmit_op.data(5) & xmit_op.data(6) & xmit_op.data(7) & '1' & '0' & xmit_op.rs;
        xmit_next_state <= XMIT_HOLD_H;
        xmit_next_cnt   <= DELAY_SETUP_HOLD;

      when XMIT_HOLD_H =>
        lcd_db4567_en_rw_rs <= xmit_op.data(4) & xmit_op.data(5) & xmit_op.data(6) & xmit_op.data(7) & '0' & '0' & xmit_op.rs;
        xmit_next_state <= XMIT_DELAY_H;
        xmit_next_cnt   <= xmit_op.delay_h;

      when XMIT_DELAY_H =>
        lcd_db4567_en_rw_rs <= X"0" & '0' & '0' & '0';
        xmit_next_state <= XMIT_SETUP_L;
        xmit_next_cnt   <= DELAY_SETUP_HOLD;

      -- xmit low nibble: setup, enable, hold, delay
      when XMIT_SETUP_L =>
        lcd_db4567_en_rw_rs <= xmit_op.data(0) & xmit_op.data(1) & xmit_op.data(2) & xmit_op.data(3)  & '0' & '0' & xmit_op.rs;
        xmit_next_state <= XMIT_ENABLE_L;
        xmit_next_cnt   <= DELAY_LCD_E;

      when XMIT_ENABLE_L =>
        lcd_db4567_en_rw_rs <= xmit_op.data(0) & xmit_op.data(1) & xmit_op.data(2) & xmit_op.data(3)  & '1' & '0' & xmit_op.rs;
        xmit_next_state <= XMIT_HOLD_L;
        xmit_next_cnt   <= DELAY_SETUP_HOLD;

      when XMIT_HOLD_L =>
        lcd_db4567_en_rw_rs <= xmit_op.data(0) & xmit_op.data(1) & xmit_op.data(2) & xmit_op.data(3)  & '0' & '0' & xmit_op.rs;
        xmit_next_state <= XMIT_DELAY_L;
        xmit_next_cnt   <= xmit_op.delay_l;

      when XMIT_DELAY_L =>
        lcd_db4567_en_rw_rs <= X"0" & '0' & '0' & '0';
        xmit_next_state <= XMIT_IDLE;
        xmit_next_cnt   <= 0;

    end case;
  end process xmit_fsm;

  status_out : process (send_state, xmit_state) is
    variable s : std_logic_vector(3 downto 0) := X"0";
    variable x : std_logic_vector(3 downto 0) := X"0";
  begin
    case send_state is
      when SEND_RESET  => s := X"1";
      when SEND_CONFIG => s := X"2";
      when SEND_READY  => s := X"3";
    end case;
    case xmit_state is
      when XMIT_IDLE     => x := X"1";
      when XMIT_SETUP_H  => x := X"2";
      when XMIT_ENABLE_H => x := X"3";
      when XMIT_HOLD_H   => x := X"4";
      when XMIT_DELAY_H  => x := X"5";
      when XMIT_SETUP_L  => x := X"6";
      when XMIT_ENABLE_L => x := X"7";
      when XMIT_HOLD_L   => x := X"8";
      when XMIT_DELAY_L  => x := X"9";
    end case;
    status <= s & x;
  end process status_out;

end architecture rtl;
