-------------------------------------------------------------------------------
-- Title      : Quad PWM motor controller
-- Project    : 
-------------------------------------------------------------------------------
-- File       : pwm4.vhdl
-- Author     : <kwalsh@holycross.edu>
-- Company    : 
-- Created    : 2016-02-07
-- Last update: 2016-02-07
-- Platform   : 
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Outputs centered PWM signals and motor direction signals.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author      Description
-- 2016-02-07  1.0      kwalsh      Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

entity pwm4 is

  port (
        clk10                 : in  std_logic;
        rst                   : in  std_logic;
        pwmcnt                : in  std_logic_vector(7 downto 0);
        power                 : in  std_logic_vector(31 downto 0); -- pwr3, pwr2, pwr1, pwr0
        en_dir                : in  std_logic_vector(31 downto 0); -- x...x, en3, dir3, en2, dir2, en1, dir1, en0, dir0
	write_power           : in  std_logic;
	write_en_dir          : in  std_logic;
	write                 : in  std_logic;
        tick                  : in  std_logic;
	dv                    : out std_logic_vector(7 downto 0) -- d3, v3, d2, v2, d1, v1, d0, v0
  );

end pwm4;

architecture rtl of pwm4 is

  function upper(x: in std_logic_vector(7 downto 0)) return std_logic_vector is
    constant max : std_logic_vector(7 downto 0) := X"FF";
    constant maxm1 : std_logic_vector(7 downto 0) := X"FE";
  begin
    if x(0) = '0' then
      return std_logic_vector(unsigned(max) -  unsigned('0' & x(7 downto 1)));
    else
      return std_logic_vector(unsigned(maxm1) - unsigned('0' & x(7 downto 1)));
    end if;
  end upper;

  function lower(x: in std_logic_vector(7 downto 0)) return std_logic_vector is
    constant one : std_logic_vector(7 downto 0) := X"01";
  begin
    if x(7) = '0' then
      return ('0' & x(7 downto 1));
    else
      return std_logic_vector(unsigned('0' & x(7 downto 1)) + unsigned(one));
    end if;
  end lower;

  function enabled(en : std_logic; cnt, A, B : std_logic_vector(7 downto 0)) return std_logic is
  begin
    if (en = '1') and ((unsigned(cnt) < unsigned(A)) or (unsigned(cnt) > unsigned(B))) then
      return '1';
    else
      return '0';
    end if;
  end enabled;

  signal enables : std_logic_vector(3 downto 0) := X"0";
  signal dirs : std_logic_vector(3 downto 0) := X"0";
  signal vs : std_logic_vector(3 downto 0) := X"0";
  signal pwr3 : std_logic_vector(7 downto 0) := X"00";
  signal pwr2 : std_logic_vector(7 downto 0) := X"00";
  signal pwr1 : std_logic_vector(7 downto 0) := X"00";
  signal pwr0 : std_logic_vector(7 downto 0) := X"00";

  signal A3 : std_logic_vector(7 downto 0) := X"00";
  signal A2 : std_logic_vector(7 downto 0) := X"00";
  signal A1 : std_logic_vector(7 downto 0) := X"00";
  signal A0 : std_logic_vector(7 downto 0) := X"00";
  signal B3 : std_logic_vector(7 downto 0) := X"00";
  signal B2 : std_logic_vector(7 downto 0) := X"00";
  signal B1 : std_logic_vector(7 downto 0) := X"00";
  signal B0 : std_logic_vector(7 downto 0) := X"00";

begin

  dv <= dirs(3) & vs(3) & dirs(2) & vs(2) & dirs(1) & vs(1) & dirs(0) & vs(0);
  -- dv <= "00011011";
  -- dv <= vs(3) & dirs(3) & vs(2) & dirs(2) & vs(1) & dirs(1) & vs(0) & dirs(0);

  iface : process (clk10) is
  begin
    if rising_edge(clk10) then
      if rst = '1' then
	enables <= X"0";
	dirs <= X"0";
	pwr3 <= X"00";
	pwr2 <= X"00";
	pwr1 <= X"00";
	pwr0 <= X"00";
      elsif write = '1' and write_power = '1' then
	pwr3 <= power(31 downto 24);
	pwr2 <= power(23 downto 16);
	pwr1 <= power(15 downto 8);
	pwr0 <= power(7 downto 0);
      elsif write = '1' and write_en_dir = '1' then
	enables <= en_dir(7) & en_dir(5) & en_dir(3) & en_dir(1);
	dirs <= en_dir(6) & en_dir(4) & en_dir(2) & en_dir(0);
      end if;
    end if;
  end process iface;

  update : process(clk10) is
  begin
    if rising_edge(clk10) then
      if rst = '1' then
	A3 <= X"00";
	A2 <= X"00";
	A1 <= X"00";
	A0 <= X"00";
	B3 <= X"00";
	B2 <= X"00";
	B1 <= X"00";
	B0 <= X"00";
      elsif tick = '1' then
	A3 <= lower(pwr3);
	B3 <= upper(pwr3);
	A2 <= lower(pwr2);
	B2 <= upper(pwr2);
	A1 <= lower(pwr1);
	B1 <= upper(pwr1);
	A0 <= lower(pwr0);
	B0 <= upper(pwr0);
      end if;
    end if;
  end process update;

  modulate : process(clk10) is
  begin
    if rising_edge(clk10) then
      if rst = '1' then
	vs <= X"0";
      else
	vs(3) <= enabled(enables(3), pwmcnt, A3, B3);
	vs(2) <= enabled(enables(2), pwmcnt, A2, B2);
	vs(1) <= enabled(enables(1), pwmcnt, A1, B1);
	vs(0) <= enabled(enables(0), pwmcnt, A0, B0);
      end if;
    end if;
  end process modulate;

end architecture rtl;
