// Author: kwalsh@holycross.edu

#include "led.h"
#include "lcd.h"
#include "joystick.h"

unsigned int joystick_axis(unsigned int js, unsigned int axis) {
  unsigned int chan = (js * 2) + axis;
  // oops, we reversed wires for channels 2 and 3...
  if (chan == 2) chan = 3;
  else if (chan == 3) chan = 2;
  unsigned int offset = (chan << 4) + 0;
  return IORD_16DIRECT(ADC_BASE, (offset << 1));
}

int joystick_baxis(unsigned int b_fwd, unsigned int b_rev, int p) {
  int f = joystick_button(0, b_fwd) ;
  int r = joystick_button(0, b_rev) ;
  if (f && !r) return p;
  else if (r && !f) return -p;
  else return 0;
}

unsigned int _jh[4], _jl[4], _jm[4]; // high, low, middle values for joysticks
float _ja[4], _jb[4]; // estimated A, B resistor values for joysticks

void joystick_demo() {
  // +--------------------+
  // |Joystick demo:      |
  // |l/r 000   l/r 000   |
  // |u/d 000   u/d 000   |
  // |btn 0765  btn 4321  |
  // +--------------------+
  lcd_printf("\fJoystick demo:\n");
  unsigned int a[4];
  unsigned int i;
  for (;;) {
    for (i = 0; i < 4; i++)
      a[i] = joystick_axis(0, i);
    unsigned int b = joystick_buttons();

    lcd_goto(1, 0);
    lcd_printf("l/r %3x   l/r %3x\n", a[0], a[2]);
    lcd_printf("u/d %3x   u/d %3x\n", a[1], a[3]);
    lcd_printf("btn %c%c%c%c  btn %c%c%c%c\n",
	(b&0x80)?'7':'_',
	(b&0x40)?'6':'_',
	(b&0x20)?'5':'_',
	(b&0x10)?'4':'_',
	(b&0x08)?'3':'_',
	(b&0x04)?'2':'_',
	(b&0x02)?'1':'_',
	(b&0x01)?'0':'_');
    usleep(100000);
  }
}

void joystick_calibrate() {
  unsigned int i, j;
  unsigned int a[4];
  
  // calibrate joystick midpoints
  // +--------------------+
  // |Calibrating: don't  |
  // |touch joysticks...  |
  // |  000   l/r   000   |
  // |  000   u/d   000   |
  // +--------------------+
  lcd_printf("\fCalibrating: don't\ntouch joysticks...\n");
  led_blink();
  
  for (i = 0; i < 4; i++)
    _jm[i] = 0;

  for (j = 0; j < 12+8; j++) {
    for (i = 0; i < 4; i++) {
      a[i] = joystick_axis(0, i);
      if (j >= 12)
	_jm[i] += a[i];
    }
    lcd_goto(2, 0);
    lcd_printf("  %3x   l/r   %3x\n", a[0], a[2]);
    lcd_printf("  %3x   u/d   %3x\n", a[1], a[3]);
    usleep(100000);
  }

  for (i = 0; i < 4; i++)
    _jm[i] /= 8;

  // calibrate joystick range
  // +--------------------+
  // |Calibrating: move   |
  // |joysticks around... |
  // |000-000 l/r 000-000 |
  // |000-000 u/d 000-000 |
  // +--------------------+
  lcd_printf("\fCalibrating: move\njoysticks around...\n");
  led_blink();

  for (i = 0; i < 4; i++)
    _jl[i] = _jh[i] = _jm[i];

  for (j = 0; j < 50; j++) {
    for (i = 0; i < 4; i++) {
      int a = joystick_axis(0, i);
      if (a < _jl[i])
	_jl[i] = a;
      if (a > _jh[i])
	_jh[i] = a;
    }
    lcd_goto(2, 0);
    lcd_printf("%3x-%3x l/r %3x-%3x\n", _jl[0], _jh[0], _jl[2], _jh[2]);
    lcd_printf("%3x-%3x u/d %3x-%3x\n", _jl[1], _jh[1], _jl[3], _jh[3]);
    usleep(100000);
  }

  lcd_printf("\fCalibration: done\n");

  calibration_configure();
}

void joystick_precalibrated() {
  alt_printf("Using preconfigured joystick data.\n");

  _jl[0] = 2; _jm[0] = 0xbea; _jh[0] = 0xdc6;
  _jl[1] = 2; _jm[1] = 0xb4a; _jh[1] = 0xdba;
  _jl[2] = 2; _jm[2] = 0xbde; _jh[2] = 0xd56;
  _jl[3] = 2; _jm[3] = 0xb48; _jh[3] = 0xd20;

  calibration_configure();
}

void calibration_configure() {
  unsigned int i;

  alt_printf("Calibration results:\n");
  for (i = 0; i < 4; i++) {
    // sanity check
    if (_jl[i] >= 0x010) {
      alt_printf("axis %x lo %x out of spec\n", i, _jl[i]);
      _jl[i] = 0;
    }
    if (_jm[i] <= 0x500 || _jm[i] >= 0xd00) {
      alt_printf("axis %x mid %x out of spec\n", i, _jm[i]);
      _jm[i] = 0xb00;
    }
    if (_jh[i] <= _jm[i] || _jh[i] >= 2*_jm[i]) {
      alt_printf("axis %x hi %x out of spec\n", i, _jh[i]);
      _jh[i] = (3 * _jm[i]) / 2;
      if (_jh[i] > 0xfff)
	_jh[i] = 0xfff;
    }

#define DEAD 96
#define MID  (255+DEAD)
#define TOP  (2*MID)
    // position scale (x) range is 0 to TOP (shifted to -255-64 to +255+64)
    // voltage scale (y) domain is 0 to 0xfff
    //   y = Ax/(B+x)
    //   x = By/(A-y)
    // The jl[] low-side voltages are usually near zero, so we'll assume they are zero.
    // The jh[] high-side values and jm[] mid-point values give us two points,
    // so we can solve for A and B.
    //   H = (A*TOP) / (B+TOP)
    //   M = (A*0x0ff) / (B+0x0ff)
    //   H*(B+TOP) = (A*TOP) 
    //   M*(B+0x0ff) = (A*0x0ff) 
    //   2*M*(B+0x0ff) = 2*(A*0x0ff) = (A*TOP) = H*(B+TOP) 
    //   2*M*B - H*B = H*TOP - 2*M*0x0ff
    //   B = (H*TOP - M*TOP) / (2*M - H)
    //   B = (H - M) * TOP / (2*M - H)
    //   A = H*(B+TOP)/TOP
    /* alt_printf("float b = ((float)(_jh[i] - _jm[i])*(float)*TOP)/((float)(2*_jm[i]) - (float)_jh[i])\n");
    alt_printf("float b = ((float)(0x%x - 0x%x)*(float)TOP)/((float)(2*0x%x) - (float)0x%x)\n",
	_jh[i], _jm[i], _jm[i], _jh[i]);	
    alt_printf("float b = ((float)(0x%x)*(float)TOP)/((float)(0x%x) - (float)0x%x)\n",
	_jh[i] - _jm[i], 2*_jm[i], _jh[i]);	
    alt_printf("float b = (0x%x.%x * 0x%x.%x)/(0x%x.%x - 0x%x.%x)\n",
	(int)((float)(_jh[i] - _jm[i])),
	((int)(((float)(_jh[i] - _jm[i])) * 0x10000))%0x10000,
	(int)((float)(TOP)),
	((int)(((float)(TOP)) * 0x10000))%0x10000,
	(int)((float)2*_jm[i]),
	((int)(((float)2*_jm[i]) * 0x10000))%0x10000,
	(int)((float)_jh[i]),
	((int)(((float)_jh[i]) * 0x10000))%0x10000);	
    alt_printf("float b = (0x%x.%x)/(0x%x.%x)\n",
	(int)((float)(_jh[i] - _jm[i])*(float)(TOP)),
	((int)(((float)(_jh[i] - _jm[i])*(float)(TOP)) * 0x10000))%0x10000,
	(int)(((float)2*_jm[i])-(float)_jh[i]),
	((int)((((float)2*_jm[i])-(float)_jh[i]) * 0x10000))%0x10000);
    alt_printf("float b = (0x%x.%x)\n",
	(int)(((float)(_jh[i] - _jm[i])*(float)(TOP))/(((float)2*_jm[i])-(float)_jh[i])),
	((int)((((float)(_jh[i] - _jm[i])*(float)(TOP))/(((float)2*_jm[i])-(float)_jh[i])) * 0x10000))%0x10000); */
    float b = ((float)(_jh[i] - _jm[i])*(float)TOP) / (float)(2*_jm[i] - _jh[i]);
    float a = (float)(_jh[i] * (b + TOP)) / (float)TOP;
    _ja[i] = a;
    _jb[i] = b;
    alt_printf("axis %x: lo %x mid %x hi %x a %x b %x\n", i, _jl[i], _jm[i], _jh[i], (int)(_ja[i]*0x1000), (int)(_jb[i]*0x1000));
  }
}

// returns number in range [-255, 255]
int joystick_pos(unsigned int js, unsigned int axis) {
  unsigned int v = joystick_axis(js, axis);
  unsigned int i = (js * 2) + axis;

  if (v < _jl[i])
    return -255;
  if (v > _jh[i])
    return 255;
  /* 
  // piecewise linear estimate
  if (v <= _jm[i]) {
    // pos = -255 * (m - v) / (m - l)
    return (-255 * (int)(_jm[i] - v)) / (int)(_jm[i] - _jl[i]);
  } else {
    // pos = 255 * (v - m) / (h - m)
    return (255 * (int)(v - _jm[i])) / (int)(_jh[i] - _jm[i]);
  }
  */
  // analytic estimate x = By/(A-y)
  float x = (_jb[i] * ((float)v)) / (_ja[i] - ((float)v));
  int xx = (int)x;
  if (xx < 0) xx = 0;
  else if (xx > TOP) xx = TOP;
  xx = xx - MID;
  // dead zone is -64 to +64
  if (xx >= DEAD) return xx - DEAD;
  if (xx <= -DEAD) return xx + DEAD;
  return 0;
}

