// Author: kwalsh@holycross.edu

#include <stdarg.h>
#include <sys/alt_stdio.h>
#include <sys/alt_sys_wrappers.h>
#include <string.h>
#include <io.h>
#include <system.h>
#include "led.h"
#include "lcd.h"

void lcd_out(unsigned int reg, unsigned char data) {
  unsigned char status;
  unsigned int count = 0;
  while ((status = IORD_8DIRECT(LCD_BASE, LCD_REG_STATUS)) != LCD_IDLE) {
    count++;
    if (count < 100) continue;
    usleep(1);
    if (count < 200) continue;
    usleep(10);
    if (count < 300) continue;
    usleep(100);
    if (count < 400) continue;
    led_blink();
    alt_printf("lcd failed with status %x\n", status);
    return;
  }

  IOWR_8DIRECT(LCD_BASE, reg, data);
}

static unsigned int _r = 0, _c = 0, _a = 0;

void lcd_putc(char c) {
  switch (c) {
    case '\n':
      lcd_goto(_r+1, 0);
      break;
    case '\r':
      lcd_goto(_r, 0);
      break;
    case '\f':
      lcd_clear();
      break;
    case '\t':
      do {
	lcd_putc(' ');
      } while (_c % 4);
      break;
    case '\v':
      lcd_goto(_r+1, _c & ~0x3);
      break;
    case '\b':
      // not yet defined
      break;
    case '\a':
      // not yet implemented
      break;
    default:
      if (_r < 4 && _c < 20) {
	lcd_dat(c);
	_a++;
      }
      _c++;
      break;
  }
}

void lcd_puts(const char *s) {
  while (*s != '\0')
    lcd_putc(*s++);
}

void lcd_clear() {
  lcd_cmd(0x01);
  usleep(1640);
  _a = 0;
  _r = _c = 0;
}

void lcd_init() {
  lcd_clear();
  lcd_cmd(0x28); // function_set(4bit, 2line, 5x8bit)
  lcd_cmd(0x06); // entry_mode(incr, noshift)
  lcd_cmd(0x0C); // display(on, nocursor, noblink)
  _a = 0;
  _r = _c = 0;
  lcd_cmd(0x80); // move to home
}

void lcd_goto(unsigned int r, unsigned int c) {
  _r = r;
  _c = c;
  unsigned int a;
  switch (r) {
    case 0: a = 0x00; break;
    case 1: a = 0x40; break;
    case 2: a = 0x14; break;
    case 3: a = 0x54; break;
    default: return;
  }
  if (c < 20) a += c;
  else return;
  if (_a == a)
    return;
  _a = a;
  lcd_cmd(0x80 | a);
}


#include <stdarg.h>
#include "sys/alt_stdio.h"

/*
 * ALT printf function
 */
  void
lcd_printf(const char* fmt, ... )
{
  va_list args;
  va_start(args, fmt);
  const char *w;
  char c;

  w = fmt;
  while ((c = *w++) != 0)
  {
    // If not a format escape character, just print it.
    if (c != '%') {
      lcd_putc(c);
      continue;
    }

    // If malformed format escape, stop.
    if ((c = *w++) == 0)
      break;

    // Handle %% escape.
    if (c == '%') {
      lcd_putc(c);
      continue;
    }

    // Handle %c escape.
    if (c == 'c') {
      int v = va_arg(args, int);
      lcd_putc(v);
      continue;
    }

    // Handle %s escape.
    if (c == 's') {
      char *s = va_arg(args, char *);
      lcd_puts(s);
      continue;
    }

    // Handle %0... escape.
    unsigned char leader = ' '; 
    if (c == '0') {
      leader = '0';
      if ((c = *w++) == 0)
	break;
    }

    // Handle %...n... escape.
    unsigned digits = 0;
    if (c >= '1' && c <= '9') {
      digits = (c - '0');
      if ((c = *w++) == 0)
	break;
    }

    // Handle %...x escape.
    if (c == 'x') {
      unsigned long v = va_arg(args, unsigned long);
      unsigned long digit;

      // Special case: print zero
      if (v == 0) {
	while (digits > 1) {
	  lcd_putc(leader);
	  digits--;
	}
	lcd_putc('0');
	continue;
      }

      int digit_shift;
      if (digits > 0) {
	digit_shift = (digits - 1) * 4;
      } else {
	// Find first non-zero digit.
	digit_shift = 28;
	while (!(0xf & (v >> digit_shift)))
	  digit_shift -= 4;
      }

      for (; digit_shift >= 0; digit_shift -= 4)
      {
	digit = 0xf & (v >> digit_shift);
	if (digit == 0)
	  lcd_putc(leader);
	else {
	  leader = '0';
	  if (digit <= 9)
	    c = '0' + digit;
	  else
	    c = 'a' + digit - 10;
	  lcd_putc(c);
	}
      }
      continue;
    }

    // if (c == 'd') ...
  }
}


